// Cjpeg.cpp
//
// JPEG Codec C++ Wrapper for IJG's JPEG C Codec
// author: Leiming Qian <qianl@us.ibm.com>
// version: 1.1 alpha
// date: 11/01/01

#include <stdio.h>
#include <stdlib.h>
#include "Cjpeg.h"

Cjpeg::Cjpeg()
{
    imgBuf = NULL;
    progressive = 0;
    restartMarkers = -1;
}

Cjpeg::~Cjpeg()
{
    if (imgBuf)
        delete [] imgBuf;
}

int Cjpeg::readJPEG(char *jpgFileName)
{
    FILE *jpgFilePtr;
    jpgFilePtr = fopen(jpgFileName, "rb");
    if (jpgFilePtr==NULL) {
        fprintf(stderr, "Cann't open jpeg file %s\n", jpgFileName);
        return(CJPEG_ERROR);
    }
    
    struct jpeg_decompress_struct dinfo;
    struct jpeg_error_mgr jerr;

    dinfo.err = jpeg_std_error(&jerr);
    jpeg_create_decompress(&dinfo);
    jpeg_stdio_src(&dinfo, jpgFilePtr);

    (void) jpeg_read_header(&dinfo, TRUE);
    (void) jpeg_start_decompress(&dinfo);
    width = dinfo.output_width;
    height = dinfo.output_height;
    colorChannel = dinfo.output_components;
    colorSpace = dinfo.out_color_space;
    frameSize = width * height;
    imgSize = frameSize * colorChannel;

    imgBuf = new JSAMPLE [imgSize];

    unsigned long rowStride = width * colorChannel;
    JSAMPARRAY decBuf = (*dinfo.mem->alloc_sarray)
        ((j_common_ptr)&dinfo, JPOOL_IMAGE, rowStride, 1);

    while (dinfo.output_scanline < height) {
        (void)jpeg_read_scanlines(&dinfo, decBuf, 1);
        memcpy( &imgBuf[(dinfo.output_scanline-1)*rowStride],
                decBuf[0], rowStride);
    }

    (void) jpeg_finish_decompress(&dinfo);
    jpeg_destroy_decompress(&dinfo);
    fclose(jpgFilePtr);
    return(CJPEG_SUCCESS);
}

unsigned int Cjpeg::getWidth() { return (width); }
unsigned int Cjpeg::getHeight() { return (height); }
unsigned int Cjpeg::getColorChannel() { return (colorChannel); }
unsigned long Cjpeg::getFrameSize() { return (frameSize); }
unsigned long Cjpeg::getImgSize() { return (imgSize); }

int Cjpeg::writeJPEG(char *jpgFileName)
{
    if (imgBuf == NULL) {
        fprintf(stderr, "image content buffer empty!\n");
        return(CJPEG_ERROR);
    }

    if ( width<=0 || height<=0 || (colorChannel!=1&&colorChannel!=3) ) {
        fprintf(stderr, "one of th image parameters are wrong\n");
        fprintf(stderr, "width = %d, height = %d, colorChannel = %d\n",
                width, height, colorChannel);
        fprintf(stderr, "run setImgSize(width, height, channel) first!\n");
        return(CJPEG_ERROR);
    }

    FILE *jpgFilePtr;
    jpgFilePtr = fopen(jpgFileName, "wb");
    if (jpgFilePtr == NULL) {
        fprintf(stderr, "Can't create output file %s\n", jpgFileName);
        return(CJPEG_ERROR);
    }
    
    struct jpeg_compress_struct cinfo;
    struct jpeg_error_mgr jerr;

    cinfo.err = jpeg_std_error(&jerr);
    jpeg_create_compress(&cinfo);
    jpeg_stdio_dest(&cinfo, jpgFilePtr);
    
    cinfo.image_width = width;
    cinfo.image_height = height;
    cinfo.input_components = colorChannel;
    cinfo.in_color_space = colorSpace;
    unsigned long rowStride = width * colorChannel;
    jpeg_set_defaults(&cinfo);
    jpeg_set_quality(&cinfo, quality, TRUE);
    if (progressive)
        jpeg_simple_progression(&cinfo);
    if (restartMarkers>=0)
        cinfo.restart_in_rows = restartMarkers;

    jpeg_start_compress(&cinfo, TRUE);
    JSAMPROW rowPtr[1];
    while (cinfo.next_scanline < cinfo.image_height) {
        rowPtr[0] = &(imgBuf[cinfo.next_scanline * rowStride]);
        jpeg_write_scanlines(&cinfo, rowPtr, 1);
    }

    jpeg_finish_compress(&cinfo);
    jpeg_destroy_compress(&cinfo);
    fclose(jpgFilePtr);

    return(CJPEG_SUCCESS);
}

void Cjpeg::setImgSize(unsigned int x, unsigned int y, unsigned int z)
{
    width = x;
    height = y;
    colorChannel = z;
    if (colorChannel == 1)
        colorSpace = JCS_GRAYSCALE;
    else
        colorSpace = JCS_RGB;
    if (imgBuf == NULL)
        imgBuf = new JSAMPLE [width*height*colorChannel];
}

void Cjpeg::setQuality(unsigned int q)
{
    if (q>0 && q<=100)
        quality = q;
    else {
        fprintf(stderr, "quality has to be between 0 and 100\n");
        fprintf(stderr, "using default quality of %d\n", DEFAULT_QUALITY);
        quality = DEFAULT_QUALITY;
    }
}

void Cjpeg::setProgressive()
{
    progressive=1;
}

void Cjpeg::setRestartMarkers(int i)
{
    if (i>0 && i<(int)height)
        restartMarkers = i;
}

int Cjpeg::exportRGB(double *R, double *G, double *B)
{
    if (imgBuf==NULL) {
        fprintf(stderr, "image buffer empty!\n");
        return(CJPEG_ERROR);
    }

    if (R==NULL || G==NULL || B==NULL) {
        fprintf(stderr, "target buffer not allocated!\n");
        return(CJPEG_ERROR);
    }

    unsigned long i;
    if (colorChannel==1) {
        // grayscale image
        for (i=0;i<width*height;i++) {
            R[i] = (double)(imgBuf[i]);
            G[i] = B[i] = R[i];
        }
    } else {
        // color RGB image
        for (i=0;i<width*height;i++) {
            R[i] = (double)(imgBuf[i*3]);
            G[i] = (double)(imgBuf[i*3+1]);
            B[i] = (double)(imgBuf[i*3+2]);
        }
    }

    return(CJPEG_SUCCESS);
}
            
int Cjpeg::importRGB(double *R, double *G, double *B)
{
    if (width<=0 || height<=0 || (colorChannel!=1 && colorChannel!=3) ) {
        fprintf(stderr, "invalid width %d height %d colorChannel %d\n",
                width, height, colorChannel);
        return(CJPEG_ERROR);
    }
    if (imgBuf==NULL)
        imgBuf = new JSAMPLE[width*height*colorChannel];

    if (R==NULL || G==NULL || B==NULL) {
        fprintf(stderr, "source buffer empty!\n");
        return(CJPEG_ERROR);
    }

    unsigned long i;
    if (colorChannel==1) {
        // grayscale image, assuming R=G=B
        for (i=0;i<width*height;i++)
            imgBuf[i] = (JSAMPLE)(R[i]);
    } else {
        // color RGB image
        for (i=0;i<width*height;i++) {
            imgBuf[i*3] = (JSAMPLE)(R[i]);
            imgBuf[i*3+1] = (JSAMPLE)(G[i]);
            imgBuf[i*3+2] = (JSAMPLE)(B[i]);
        }
    }

    return(CJPEG_SUCCESS);
}
